#!/usr/bin/perl -w

use strict;
use Fcntl qw(:seek);
use IPC::SysV qw(IPC_PRIVATE S_IRWXU IPC_CREAT);
use IPC::Semaphore;
use POSIX qw(:sys_wait_h);
use IO::Handle qw(autoflush);
use RNA;
use lib qw(../lib/);
use BAOBAB::RNA2D;

my $nchildren = 0;
my $nspawned  = 0;

sub reaper {
	while ( waitpid( -1, WNOHANG ) > 0 ) {
		$nchildren--;
	}
	$SIG{CHLD} = \&reaper;
}

use Data::Dumper qw(&Dumper);

unless ( @ARGV >= 3 ) {
	die "Usage: $0 genome reglist outfile [nmax [startat [stopat]]]\n";
}

my $fname   = shift;
my $finame  = $fname . ".idx";
my $rfname  = shift;
my $ofname  = shift;
my $nmax    = shift;
my $startat = shift;
my $stopat  = shift;
$startat = 0 unless defined $startat;

open( STDERR, ">$$.err" ) or die "Cannot redirect stderr : $!\n";

if ( !-e $finame ) {
	warn "Computing index...\n";
	my @seq = compute_index($fname);
	my $idx;
	open( $idx, ">$finame" ) or die "Cannot open $finame : $!\n";
	print $idx $_->[0] . ":" . $_->[1] . "\n" for @seq;
	close($idx);
}

my @index = load_index($finame);
warn "Loaded " . scalar(@index) . " entries\n";

my ( $file, $rfile, $ofile );

open( $file,  $fname )     or die "Cannot open $fname : $!\n";
open( $rfile, $rfname )    or die "Cannot open $rfname : $!\n";
open( $ofile, ">$ofname" ) or die "Cannot open $ofname : $! \n";

autoflush( $ofile, 1 );

my ($cloc) = (0);

my @pids = ();
$nmax = $nmax || 250;

my $sem = new IPC::Semaphore( IPC_PRIVATE, 1, S_IRWXU | IPC_CREAT );
$sem->setval( 0, 0 );

$SIG{CHLD} = \&reaper;

while (<$rfile>) {
	chomp;
	my @f = split /:/;
	my ( $seq, $strand, $offset, $stop ) = @f[ 0, 1, 2, 3 ];

	next if $seq < $startat;
	last if defined $stopat and $seq > $stopat;

	my $region = get_region(
		file     => $file,
		sequence => $seq,
		offset   => $offset,
		len      => $stop - $offset + 1,
		idx      => \@index
	);

	$region = reverse_complement($region) if $strand eq 'R';

	my $pid = fork();

	unless ( defined $pid ) {
	#	speak("Could not fork() any more processes : $!");
		die "Could not fork() any more processes : $!\n";
	}

	if ( $pid != 0 ) {
		$nchildren++;
		$nspawned++;
		sleep 1 while $nchildren > $nmax;
	}
	else {

		$region =~ y/Tt/Uu/;
		my ( $structure, $mfe ) = RNA::fold($region);
		my $rna =
		  BAOBAB::RNA2D->new( structure => $structure, sequence => $region );

		my $sl = $rna->largest_stemloop;

		if ( defined $sl ) {
			my $crop = substr $region, $sl->{start}, $sl->{len};
			( $structure, $mfe ) = RNA::fold($crop);

			if ( $strand eq 'F' ) {
				$cloc = $offset + $sl->{start};
			}
			else {

				# starts here from the point of view
				# of other scripts
				# which add something to this offset
				$cloc = $stop - $sl->{stop};
			}

			#SEQ;COORD:MFE:STEM:LEFT:RIGHT:LOOP;

			die "Semaphore trouble @ $seq:$strand:$offset : $!\n"
			  unless $sem->op( 0, 0, 0, 0, 1, 0 );
			{

				# print "[$$] reached the critical zone\n";

				# right/left arm lenghts are swapped for reverse
				# strand stem loops to allow for simpler
				# handling in successive scripts

				print {$ofile} join(
					":",
					(
						$seq,
						$strand,
						$cloc,
						sprintf( "%.2f", $mfe ),
						$sl->{stem_len},
						$strand eq 'F' ? $sl->{left_arm_len}
						: $sl->{right_arm_len},
						$strand eq 'F' ? $sl->{right_arm_len}
						: $sl->{left_arm_len},
						$sl->{loop_len}
					)
				) . "\n";

			}

			die "Semaphore trouble @ $seq:$strand:$offset : $!\n"
			  unless $sem->op( 0, -1, 0 );

			# print "-";

		}
		exit(0);

	}

}

# wait for final children to finish
sleep 1 while $nchildren > 0;

# speak("I have finished processing!");

close($ofile);
close($rfile);
close($file);

print "SPAWNED = $nspawned\n";

sub speak {
	my $text = shift;
	if ( $^O eq 'darwin' ) {

		#use Mac::Speech;
		#my $speaker = "Bruce";
		#my $voice = $Voice{$speaker};
		#my $channel = NewSpeechChannel($voice) or die $^E;
		#SpeakText($channel, $text) or die $^E;
		#while (SpeechBusy()) {};
		#DisposeSpeechChannel($channel);
	}
}

sub hamming_distance {
	my ( $s1, $s2 ) = ( shift, shift );

	my $l = ( sort ( length $s1, length $s2 ) )[0];
	my $c = 0;
	for my $i ( 0 .. $l - 1 ) {
		$c++ if substr( $s1, $i, 1 ) ne substr( $s2, $i, 1 );
	}
	return $c;
}

sub reverse_complement {
	my $seq = shift;
	my @seq = split //, $seq;
	@seq = reverse(@seq);
	$seq = join( '', @seq );
	$seq =~ y/ATGCatgc/TACGtacg/;
	return $seq;
}

sub get_region {
	no strict 'subs';
	my %a      = @_;
	my $idx    = $a{idx};
	my $len    = $a{len};
	my $offset = $a{offset};
	my $seq    = $a{sequence};
	my $file   = $a{file};

	my $b = $idx->[$seq];
	my ( $start, $stop ) = @$b;
	if ( $offset > $stop - $start ) {
		die "Offset too big\n";
	}
	seek( $file, $start,  SEEK_SET );
	seek( $file, $offset, SEEK_CUR );
	my $region;
	read( $file, $region, $len );
	return $region;
}

sub load_index {
	my @index  = ();
	my $finame = shift;
	my $idx;

	open( $idx, $finame ) or die "Cannot open $finame : $!\n";
	while (<$idx>) {
		chomp;
		my ( $start, $stop ) = split /:/;
		push @index, [ $start, $stop ];
	}
	return wantarray ? @index : \@index;
}

sub compute_index {
	my $fname = shift;
	my ($file);
	open( $file, $fname ) or die "Cannot read $fname : $!\n";

	my $state = 0;
	my @sequences;
	my ( $start, $stop );
	while ( !eof($file) ) {
		my ($c);
		read( $file, $c, 1 );
		if ( $c eq '>' ) {
			if ( $state == 3 ) {    # No \n between seqs
				push @sequences, [ $start, $stop ];
			}
			$state = 1;
		}
		elsif ( $state == 1 ) {
			if ( $c eq "\n" ) {
				$state = 2;
				$start = tell($file);
			}
		}
		elsif ( $state == 2 && $c eq "\n" ) {
			$state = 3;
			$stop  = tell($file);    #just in case eof follows
		}
		elsif ( $state == 3 && $c eq "\n" ) {
			$state = 0;
			$stop  = tell($file);
			push @sequences, [ $start, $stop ];
		}
		elsif ( $state == 3 && $c ne "\n" ) {
			$state = 2;
		}
	}

	# in case there is no empty line at the end
	if ( defined $start && defined $stop ) {
		push @sequences, [ $start, $stop ];
	}

	close($file);

	return wantarray ? @sequences : \@sequences;

}

