/*

 Extractor entry point

 by Nuno D. Mendes
 $Id: extractor.c,v 1.31 2009/12/10 12:53:17 rnl14480 Exp $

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "util.h"
#include "extractor.h"
#include "debug.h"
#include "clo.h"
#include "sequences.h"
#include "fasta.h"
#include "lalign.h"
#include "hit.h"
#include "hlist.h"
#include "types.h"

stat_t stat = { 0, 0 };

clo_t clo[] = { { 'w', "window", "Size of sliding window", 1, 0, 0, "" }, {
		'o', "output-file", "Output filename", 1, 0, 0, "" }, { 's',
		"sequence", "Process only this sequence from the input file", 1, 0, 0,
		"" }, { 'h', "help", "Print this message", 0, 1, 0, "" } };

int main(int argc, char **argv) {
	int ret = 0, left_arg = 0, seq = 0, window = 0;
	int slen = 0, pwindow = 0, tlen = 0, singleseq = 0;
	char *option = NULL, *s = NULL;
	FILE *fasta = NULL, *fout = NULL;
	dataset_t *ds = NULL;
	model_t *fmodel = NULL, *rmodel = NULL;
	align_t *falign = NULL, *ralign = NULL;
	hit_t *fhit = NULL, *rhit = NULL;
	hlist_t *fhlist = creat_hlist(), *rhlist = creat_hlist();
	float flastnzdev = 0, rlastnzdev = 0, fcurdev = 0, rcurdev = 0, fcurscore =
			0, rcurscore = 0;
	/* uint		offset = 0, poffset = 0; */

	ret = parse_clo(clo, argc, argv, &left_arg, &option);
	if (ret == CLO_TOOFEW) {
		message("Not enough arguments");
		usage(clo);
		exit(-1);
	} else if (ret == CLO_MISSINGPARAM || ret == CLO_BADFORMAT) {
		message("One option has a mandatory parameter");
		usage(clo);
		exit(-1);
	} else if (ret == CLO_UNKNOWN_OPTION) {
		message("Unknown option : %s", option);
		usage(clo);
		exit(-1);
	}
	if (clo_is_set(clo, 'h')) {
		usage(clo);
		exit(0);
	}
	if (clo_is_set(clo, 'w'))
		pwindow = atoi(clo_param(clo, 'w'));
	else
		pwindow = DEFAULT_WINDOWSIZE;
	if (clo_is_set(clo, 'o'))
		fout = safe_fopen(clo_param(clo, 'o'), "w");
	else
		fout = stdout;
	if (clo_is_set(clo, 's'))
		singleseq = atoi(clo_param(clo, 's'));

	message("There %d arguments left to parse", left_arg);

	if (left_arg) {
		message("Opening file on argv[%d]", argc - left_arg);
		fasta = safe_fopen(argv[argc - left_arg], "r");
		message("Parsing fasta file");
		ds = parse_fasta(fasta);
		message("Done.");
		fclose(fasta);
	} else {
		ds = parse_fasta(stdin);
	}

	message("Inserted window size is %d", pwindow);
	message("Dataset has %d sequence", ds->seqn);

	fmodel = creat_model(1, 1, 1, S_FORWARD);
	rmodel = creat_model(1, 1, 1, S_REVERSE);

	for (seq = 0; seq < ds->seqn; seq++) {
		slen = strlen(ds->sequences[seq]);
		tlen += slen;
		message("[Seq %d, len=%d]", seq, slen);
		if ((ds->sequences[seq])[slen] != '\0') {
			message("seq %d is not a ZSTRING", seq);
			exit(0);
		}
	}
	message("total dataset length = %d", tlen);
	if (singleseq != 0)
		message("Single seq mode ON. Processing only seq #%d", singleseq - 1);

	for (seq = 0; seq < ds->seqn; seq++) {
		if (singleseq && singleseq - 1 != seq)
			continue;
		window = pwindow;
		s = ds->sequences[seq];
		slen = strlen(s);
		message("Dealing with seq[%d,len=%d]", seq, slen);
		if (slen <= 2)
			continue;
		if (slen < window)
			window = slen;

		/* DEBUG assertion */
		/*
		 assert(hlist_nhits(fhlist) == 0);
		 assert(hlist_nhits(rhlist) == 0);
		 */

		while (*(s + window - 1) != '\0') {
			/*
			 offset = s - ds->sequences[seq];
			 if (offset != 0)
			 assert(offset - 1 == poffset);
			 poffset = offset;
			 */
			falign = compute_match(s, window, fmodel);
			ralign = compute_match(s, window, rmodel);

			fhit = creat_hit(falign, window, s - ds->sequences[seq], seq,
					S_FORWARD);
			rhit = creat_hit(ralign, window, s - ds->sequences[seq], seq,
					S_REVERSE);

			process_alignment(fhit, fhlist, &fcurscore, &flastnzdev, &fcurdev,
					fout);

			process_alignment(rhit, rhlist, &rcurscore, &rlastnzdev, &rcurdev,
					fout);

			/*
			 message("fhlist = %d", hlist_nhits(fhlist));
			 message("rhlist = %d", hlist_nhits(rhlist));
			 */

			s++;
		}

		/* process remaining hits */
		flush_remainder_hlist(fhlist, flastnzdev, fout);
		flush_remainder_hlist(rhlist, rlastnzdev, fout);

	}

	/* Cleanup */
	destroy_model(fmodel);
	destroy_model(rmodel);
	hlist_destroy(fhlist);
	hlist_destroy(rhlist);
	destroy_dataset(ds);

	message("** Statistics **");
	message("Forward strand: (+) %d; (-) %d", stat.fpositive, stat.fnegative);
	message("Reverse strand: (+) %d; (-) %d", stat.rpositive, stat.rnegative);

	if (fout != stdout)
		fclose(fout);

	return 0;
}

void process_alignment(hit_t * hit, hlist_t * hl, float *curscore,
		float *lastnzdev, float *curdev, FILE * fout) {

	float prevscore = 0.0;

	prevscore = *curscore;
	*curscore = hit_hitscore(hit);
	*curdev = *curscore - prevscore;

	if (*curdev != 0) {
		*lastnzdev = *curdev;
	}
	if (*curdev < 0) {
		/* flush with 1 */
		flush_region(hl, fout);
		/* discard with 0 */
		discard_hit(hit);

	} else if (*curdev == 0) {
		if (*lastnzdev > 0) {
			/* add to hlist */
			hl = hlist_add_hit(hl, hit);
		} else {
			/* discard */
			discard_hit(hit);
		}
	} else if (*curdev > 0) {
		/* discard with 0 */
		discard_region(hl);

		/* add to hlist */
		hl = hlist_add_hit(hl, hit);

	}

	return;
}

void flush_region(hlist_t * hl, FILE * fout) {
	hit_t *h = NULL;
	hiterator_t *hi = NULL;
	unsigned long start = 0, stop = 0;
	uint sequence = 0;
	strand_t strand;
	float hitscore = 0.0;

	if (hlist_nhits(hl) == 0)
		return;

	h = hlist_get_first(hl);
	start = h->offset;
	stop = h->offset + h->window - 1;
	sequence = h->sequence;
	strand = h->strand;
	hitscore = hit_hitscore(h);

	hi = hlist_iterate_reset(hl);
	while (hlist_iterate_has_next(hi)) {
		h = hlist_iterate_next(hi);
		stop = h->offset + h->window - 1;
		if (strand == S_FORWARD) {
			stat.fpositive++;
		} else {
			stat.rpositive++;
		}
		hlist_remove_hit(hl, h);
		destroy_hit(h);
	}
	hlist_iterate_finish(hi);

	fprintf(fout, "%d:%c:%ld:%ld:%.3f\n", sequence, strand == S_FORWARD ? 'F'
			: 'R', start, stop, hitscore);
	return;

}

void discard_hit(hit_t * hit) {
	if (hit == NULL)
		return;

	if (hit->strand == S_FORWARD) {
		stat.fnegative++;
	} else {
		stat.rnegative++;
	}
	destroy_hit(hit);
}

void discard_region(hlist_t * hl) {
	hiterator_t *hi = NULL;
	hit_t *h = NULL;

	hi = hlist_iterate_reset(hl);
	while (hlist_iterate_has_next(hi)) {
		h = hlist_iterate_next(hi);
		if (h->strand == S_FORWARD) {
			stat.fnegative++;
		} else {
			stat.rnegative++;
		}
		hlist_remove_hit(hl, h);
		destroy_hit(h);
	}
	hlist_iterate_finish(hi);
}

void flush_remainder_hlist(hlist_t * hlist, float lastnzdev, FILE * fout) {
	int flush_with = lastnzdev > 0 ? 1 : 0;

	if (flush_with == 1) {
		flush_region(hlist, fout);
	} else {
		discard_region(hlist);
	}

	return;
}
