/*
 * /
 *
 * Functions for performing a Smith-Waterman alignment
 *
 * by Nuno D. Mendes $Id: lalign.c,v 1.9 2009/06/13 17:55:19 rnl14480 Exp $
 *
 *
 */

#include "lalign.h"
#include "util.h"
#include "debug.h"
#include <stdio.h>
#include <string.h>

#define LSTR(s,i,w) (*(s+i-1))
#define RSTR(s,i,w) (*(s+(w-1)-(i-1)))

align_t *compute_match(char *s, int window, model_t * model) {
	unsigned int n = window / 2;
	float **m = NULL;
	dir_t **t = NULL;
	align_t *align = NULL;
	int i = 0, j = 0, mlen = 0;

	float maxv = 0, curv = 0, mscore = 0;
	int maxi = 0, maxj = 0;

	if (s == NULL || model == NULL)
		return NULL;

	m = (float **) safe_malloc(sizeof(float *) * (n + 1));
	for (i = 0; i <= n; i++) {
		m[i] = (float *) safe_malloc(sizeof(float) * (n + 1));
		m[i][0] = 0;
	}
	for (j = 0; j <= n; j++)
		m[0][j] = 0;

	t = (dir_t **) safe_malloc(sizeof(dir_t *) * (n + 1));
	for (i = 0; i <= n; i++) {
		t[i] = (dir_t *) safe_malloc(sizeof(dir_t) * (n + 1));
		t[i][0] = D_ANY;
	}
	for (j = 0; j <= n; j++)
		t[0][j] = D_ANY;

	/* SMITH-WATERMAN */

	align = (align_t *) safe_malloc(sizeof(align_t));
	align->align_string = (dir_t *) safe_malloc(sizeof(dir_t) * (n + n + 1));

	for (i = 1; i <= n; i++) {
		for (j = 1; j <= n; j++) {

			mscore
					= model->score_matrix[(int) LSTR(s, i, window)][(int) RSTR(s, j, window)];

			curv = m[i - 1][j - 1] + mscore;

			if (m[i][j - 1] + model->gap > curv) {
				t[i][j] = D_UP;
				m[i][j] = m[i][j - 1] + model->gap;
			} else if (m[i - 1][j] + model->gap > curv) {
				t[i][j] = D_LEFT;
				m[i][j] = m[i - 1][j] + model->gap;
			} else {
				m[i][j] = m[i - 1][j - 1] + mscore;
				if (mscore > 0)
					t[i][j] = D_MATCH;
				else
					t[i][j] = D_MISMATCH;
			}

			if (m[i][j] < 0)
				m[i][j] = 0;

			if (m[i][j] > maxv) {
				maxi = i;
				maxj = j;
				maxv = m[i][j];
			}
		}
	}

	/* backtrack */
	i = maxi;
	j = maxj;

	align->max_i = i;
	align->max_j = j;

	while (m[i][j] > 0) {
		align->align_string[mlen++] = t[i][j];

		if (t[i][j] == D_LEFT) {
			i--;
		} else if (t[i][j] == D_UP) {
			j--;
		} else {
			i--;
			j--;
		}

		align->i = i;
		align->j = j;
	}
	align->score = maxv;
	align->mlen = mlen;

	for (i = 0; i <= n; i++)
		safe_free(m[i]);
	safe_free(m);

	for (i = 0; i <= n; i++)
		safe_free(t[i]);
	safe_free(t);

	/*
	 print_alignment(s, window, align);
	 */

	return align;

}

void print_alignment(char *s, int window, align_t * align) {
	int i = 0;
	int curi = 0, curj = 0;

	if (s == NULL || align == NULL)
		return;

	curi = align->i + 1;
	curj = align->j + 1;

	for (i = 0; i < align->mlen; i++)
		switch (align->align_string[i]) {
		case D_MATCH:
			printf("M");
			break;
		case D_MISMATCH:
			printf("m");
			break;
		case D_UP:
			printf("U");
			break;
		case D_LEFT:
			printf("L");
			break;
		case D_ANY:
			printf(" ");
			break;
		default:
			;
		}
	printf("\n");

	for (i = align->mlen; i > 0; i--)
		switch (align->align_string[i - 1]) {
		case D_MATCH:
			printf("*");
			break;
		case D_MISMATCH:
		case D_UP:
		case D_LEFT:
		case D_ANY:
			printf(" ");
			break;
		default:
			;
		}
	printf("\n");

	for (i = align->mlen; i > 0; i--)
		switch (align->align_string[i - 1]) {
		case D_MATCH:
		case D_MISMATCH:
		case D_LEFT:
		case D_ANY:
			printf("%c", LSTR(s, curi++, window));
			break;
		case D_UP:
			printf("-");
			break;
		default:
			;
		}

	printf("\n");

	for (i = align->mlen; i > 0; i--)
		switch (align->align_string[i - 1]) {
		case D_MATCH:
		case D_MISMATCH:
		case D_UP:
		case D_ANY:
			printf("%c", RSTR(s, curj++, window));
			break;
		case D_LEFT:
			printf("-");
			break;
		default:
			;

		}

	printf("\n");

}

int align_score(align_t * align) {
	if (align == NULL)
		return 0;
	return align->score;
}

align_t *align_clone(align_t * align) {
	align_t *n = NULL;
	int i = 0;
	if (align == NULL)
		return n;

	n = (align_t *) safe_malloc(sizeof(align_t));

	n->i = align->i;
	n->j = align->j;
	n->max_i = align->max_i;
	n->max_j = align->max_j;
	n->mlen = align->mlen;
	n->score = align->score;

	n->align_string = (dir_t *) safe_malloc(sizeof(dir_t) * align->mlen);
	for (i = 0; i < align->mlen; i++)
		n->align_string[i] = align->align_string[i];

	return n;

}

model_t *creat_model(float a0, float a1, float a2, strand_t s) {
	/* creat_model(float, float, float, FORWARD/REVERSE) */
	model_t *model = NULL;
	int i = 0, j = 0;
	float **matrix = NULL;

	model = (model_t *) safe_malloc(sizeof(model_t));
	model->gap = -a2;

	matrix = (float **) safe_malloc(sizeof(float *) * 256);

	for (i = 0; i < 256; i++) {
		matrix[i] = (float *) safe_malloc(sizeof(float) * 256);
		for (j = 0; j < 256; j++) {
			matrix[i][j] = -a1;
		}
	}

	matrix['a']['t'] = a0;
	matrix['t']['a'] = a0;
	matrix['A']['T'] = a0;
	matrix['T']['A'] = a0;
	matrix['a']['u'] = a0;
	matrix['u']['a'] = a0;
	matrix['A']['U'] = a0;
	matrix['U']['A'] = a0;

	matrix['c']['g'] = a0;
	matrix['g']['c'] = a0;
	matrix['C']['G'] = a0;
	matrix['G']['C'] = a0;

	if (s == S_FORWARD) {
		matrix['g']['t'] = a0;
		matrix['t']['g'] = a0;
		matrix['G']['T'] = a0;
		matrix['T']['G'] = a0;
		matrix['g']['u'] = a0;
		matrix['u']['g'] = a0;
		matrix['G']['U'] = a0;
		matrix['U']['G'] = a0;

	} else {
		matrix['c']['a'] = a0;
		matrix['a']['c'] = a0;
		matrix['C']['A'] = a0;
		matrix['A']['C'] = a0;
	}

	model->score_matrix = matrix;

	return model;
}

void destroy_model(model_t * model) {
	int i = 0;

	if (model == NULL)
		return;

	for (i = 0; i < 256; i++)
		safe_free(model->score_matrix[i]);

	safe_free(model->score_matrix);
	safe_free(model);
}

void destroy_align(align_t * align) {
	if (align == NULL)
		return;

	safe_free(align->align_string);
	safe_free(align);
}
